/*
Simple RTSP client.
*/
#include <string.h>// strlen, strstr, memset
#include <stdio.h>// fprintf, snprintf
#include <fcntl.h>// fcntl
#include <ctype.h>// isgraph
#include <stdlib.h>// atoi, free
//#include <stdint.h>// uint16_t, uint32_t...
#include <time.h>// time, time_t

#include "rtsp.h"
#include "common.h"
#include "communication.h"

#define RTSP_DEFAULT_PORT 554
#define RTSP_ALT_PORT    8554
/*
In typical scenario, the total packet size is 1,500 bytes,
which includes the following headers and payloads: 
- IP header = 20 bytes  
- UDP header = 8 bytes 
- RTP header = 12 bytes 
- RTP payload format header = 4 bytes 
- ASF data packet header = 12 bytes 
- ASF payload headers and compressed media data = 1,444 bytes  
*/
#define RTSP_HEADER_LEN   8
#define MMS_HEADER_LEN   12

#define MAX_SESSION_ID_LENGTH 31// usially uint32, i.e. max 10 digits + zero
static char web_session_id_str[MAX_SESSION_ID_LENGTH + 1];

static unsigned int web_cseq;// uint32
static time_t web_timeout;

/* Init http header parameters */
static void init_parameters()
{
    web_session_id_str[0] = '\0';
    com_content_length = 0;
}

/** @return -1 - error, 0 - success */
static int rtsp_header_line_parser()
{
    int tmp;
    size_t pos = 0;
    if( strnicmp(&web_buf[pos], "Content-Length:", strlen("Content-Length:")) == 0 )
    {
        pos += strlen("Content-Length:");
        while( web_buf[pos] == ' ' ) pos++;// skip spaces
        tmp = atoi( &web_buf[pos] );
        if( tmp < 0 )
            DPRINT("Bad header: %s\n", &web_buf[pos]);
        else
            com_content_length = (size_t) tmp;
    } else if( strnicmp(&web_buf[pos], "Session:", strlen("Session:")) == 0 )
    {// reading parameter "Session:". Example: "Session: 14899032063100302743;timeout=60"
        tmp = 0;
        pos += strlen("Session:");
        while( web_buf[pos] == ' ' ) pos++;// skip spaces
        while
        (
            web_buf[pos] != 0 &&
            isgraph( web_buf[pos] ) &&
            web_buf[pos] != ';' &&
            tmp < MAX_SESSION_ID_LENGTH
        )
        {
            web_session_id_str[tmp++] = web_buf[pos++];
        }
        web_session_id_str[tmp] = '\0';
        if( sscanf( &web_buf[pos], "; timeout = %d", &tmp ) == 1 )
        {// minimum allowed value is 10 s, maximum value is 4294967 296 ms
            if( tmp < 10 || tmp > 4294967 )
                DPRINT("Bad timeout: %s\n", &web_buf[pos]);
            else
                web_timeout = tmp;
        }
    }
    return 0;
}
/** Base64 conversation table */
static const unsigned char ascii_table[128] =
{
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0x3e, 0xff, 0xff, 0xff, 0x3f,
    0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b,
    0x3c, 0x3d, 0xff, 0xff, 0xff, 0xf1, 0xff, 0xff,// 0xf1 for '='
    0xff, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,
    0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,
    0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16,
    0x17, 0x18, 0x19, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20,
    0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28,
    0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30,
    0x31, 0x32, 0x33, 0xff, 0xff, 0xff, 0xff, 0xff
};

static inline unsigned char to_ascii(char c)
{
    if( (unsigned char)c > 127 ) return 0xff;
    return ascii_table[(size_t)c];
}
/** Decode from base64 data. @return dst - start of the decoded data. @return length of the data */
static size_t decode_base64(char *dst, const char *src, size_t len)
{
    size_t i, pos, written;
    unsigned char c, buf[4];

    pos = 0;
    written = 0;
    //skip none-base64 characters
    while( (to_ascii(*src) == 0xff) && (pos < len) )
    {
        src++;
        pos++;
    }

    while( 1 )
    {
        for( i = 0; (i <= 3) && (pos < len); src++, pos++ )
        {
            c = to_ascii( *src );
            if( c == 0xff )
                break;
            buf[i++] = c;
        }
        if( i <= 3 )
            break;
        written += 3;
        *dst++ = (buf[0] << 2) | (buf[1] >> 4);
        *dst++ = (buf[1] << 4) | (buf[2] >> 2);
        *dst++ = (buf[2] << 6) |  buf[3];
        if( buf[2] == 0xf1 )// '='
            written--;
        if( buf[3] == 0xf1 )// '='
            written--;
    }
    return written;
}

/** @param pos - position to read data in the buffer, @return -1 - error, 0 - success */
static int rtsp_content_line_parser()
{
    size_t pos = 0;
    if( strnicmp(&web_buf[pos],
                "a=pgmpu:data:application/vnd.ms.wms-hdr.asfv1;",
                strlen("a=pgmpu:data:application/vnd.ms.wms-hdr.asfv1;")) == 0 )
    {
        pos += strlen("a=pgmpu:data:application/vnd.ms.wms-hdr.asfv1;");
        while( web_buf[pos] == ' ' ) pos++;// skip spaces
        if( strnicmp(&web_buf[pos], "base64,", strlen("base64,")) == 0 )
        {
            pos += strlen("base64,");
            return out_sound_stream
                (
                    decode_base64( &web_buf[0], &web_buf[pos], strlen( &web_buf[pos] ) )
                );
        }
        WEB_MSG("Unknown WMA header encoding type:\n%s\n", &web_buf[pos] );
        return -1;
    }
    return 0;
}
/** Process rtsp content. @return 0 if success, -1 otherwise. */
static int read_content( int(*line_parser)() )
{
    size_t readed;
    int len;
    readed = 0;
    do
    {
        len = read_socket_line( 0 );
        if( len < 0 )
            return -1;
        if( line_parser() < 0 )
            return -1;
        readed += (size_t) len;
    }
    while( readed < com_content_length );
    return 0;
}
/** Send TEARDOWN */
static void close_session()
{
    int n;
#if 0
#define RTSP_PAUSE_STR "PAUSE rtsp://%s/%s RTSP/1.0\r\n" \
                    "CSeq: %d\r\n" \
                    "User-Agent: "APP_NAME"\r\n" \
                    "Session: %s\r\n" \
                    "\r\n"
    snprintf( web_buf, BUF_SIZE - 1, RTSP_PAUSE_STR,
                       web_host, web_file, ++web_cseq, web_session_id_str );
    DPRINT( web_buf );
    if( send_data( web_buf, n ) >= 0 )
    {
        init_parameters();
        n = read_response( rtsp_header_line_parser );
        if( n >= 0 && com_content_length > 0 )
            skip_content( (size_t) n );
    }
    /** FIXME send content */
#define RTSP_LOGPLAY_STR "SET_PARAMETER rtsp://%s/%s RTSP/1.0\r\n" \
                    "CSeq: %d\r\n" \
                    "User-Agent: "APP_NAME"\r\n" \
                    "Session: %s\r\n" \
                    "Content-Length: %d\r\n" \
                    "Content-Type: application/x-wms-Logplaystats;charset=UTF-8\r\n" \
                    "\r\n"
    snprintf( web_buf, BUF_SIZE - 1, RTSP_LOGPLAY_STR,
                       web_host, web_file, ++web_cseq, web_session_id_str, 0 );
    DPRINT( web_buf );
    if( send_data( web_buf, n ) >= 0 )
    {
        init_parameters();
        n = read_response( rtsp_header_line_parser );
        if( n < 0 || com_content_length > 0 )
            return -1;// response cannot have a message body
    }
#endif
#define RTSP_TEARDOWN_STR "TEARDOWN rtsp://%s/%s RTSP/1.0\r\n" \
                    "CSeq: %d\r\n" \
                    "User-Agent: "APP_NAME"\r\n" \
                    "Session: %s\r\n" \
                    "\r\n"
    n = snprintf( web_buf, BUF_SIZE - 1, RTSP_TEARDOWN_STR,
                           web_host, web_file, ++web_cseq, web_session_id_str );

}
/** Describe on the stream. @return read position in the buffer or -1 if an error */
static int describe()
{
    int n;
    /* Sending DESCRIBE */
    DPRINT("Sending DESCRIBE request\n");
#define RTSP_DESCRIBE_STR "DESCRIBE rtsp://%s/%s RTSP/1.0\r\n" \
                    "CSeq: %i\r\n" \
                    "User-Agent: "APP_NAME"\r\n" \
                    "Accept: application/sdp\r\n" \
                    "\r\n"
    n = snprintf( web_buf, BUF_SIZE - 1, RTSP_DESCRIBE_STR, web_host, web_file, ++web_cseq );

    DPRINT( web_buf );
    if( send_data( web_buf, n ) < 0 )
        return -1;

    /* Reading response */
    DPRINT("Reading response\n");
    init_parameters();
    n = read_response( rtsp_header_line_parser );
    if( n < 0 )
        return -1;
    if( com_content_length > 0 && read_content( rtsp_content_line_parser ) < 0 )
        return -1;

    /* Sending SETUP */
    DPRINT("Sending SETUP request\n");
#define RTSP_SETUP_STR "SETUP rtsp://%s/%s/audio RTSP/1.0\r\n" \
                    "CSeq: %i\r\n" \
                    "User-Agent: "APP_NAME"\r\n" \
                    "Transport: RTP/AVP/TCP;unicast;interleaved=0-1;mode=PLAY\r\n" \
                    "\r\n"
    n = snprintf( web_buf, BUF_SIZE - 1, RTSP_SETUP_STR, web_host, web_file, ++web_cseq );

    DPRINT( web_buf );
    if( send_data( web_buf, n ) < 0 )
        return -1;

    /* Reading response */
    DPRINT("Reading response\n");
    init_parameters();
    n = read_response( rtsp_header_line_parser );
    if( n < 0 )
        return -1;
    if( com_content_length > 0 && skip_content() < 0 )
        return -1;

    /* Sending PLAY request */
    DPRINT("Sending PLAY request\n");
#define RTSP_PLAY_STR "PLAY rtsp://%s/%s RTSP/1.0\r\n" \
                    "CSeq: %i\r\n" \
                    "User-Agent: "APP_NAME"\r\n" \
                    "Session: %s\r\n" \
                    "Range: npt=0.000-\r\n" \
                    "\r\n"
    n = snprintf( web_buf, BUF_SIZE - 1, RTSP_PLAY_STR,
                                    web_host, web_file, ++web_cseq, web_session_id_str );

    DPRINT( web_buf );
    if( send_data( web_buf, n ) < 0 )
        return -1;

    /* Reading response */
    DPRINT("Reading response\n");
    init_parameters();
    n = read_response( rtsp_header_line_parser );
    if( n >= 0 && com_content_length > 0 )
        n = skip_content();
    return n;
}
/** Send Keep-Alive request to the server. @return 0 if success, -1 if an error. */
static int send_keep_alive()
{
    int n;
    /* Sending GET_PARAMETER request */
    DPRINT("Sending KeepAlive as GET_PARAMETER request\n");
#define RTSP_GET_PARAMETER_STR "GET_PARAMETER rtsp://%s/%s RTSP/1.0\r\n" \
                    "CSeq: %i\r\n" \
                    "User-Agent: "APP_NAME"\r\n" \
                    "Session: %s\r\n" \
                    "\r\n"
    n = snprintf( web_buf, BUF_SIZE - 1, RTSP_GET_PARAMETER_STR,
                web_host, web_file, ++web_cseq, web_session_id_str );
    DPRINT( web_buf );
    return send_data( web_buf, n );
}

static int play()
{
    size_t size;// uint16
    size_t block;
    size_t read_pos;
    size_t n = 0;
    time_t curr_time;
    time_t prev_time;

    prev_time = time( NULL );
    /* Prebuffering */
    DPRINT("Start prebuffering\n");
    if( prebuffering() < 0 ) return -1;
    WEB_MSG("Prebuffering complete, start playing\n");
    /* Playing */
    while( !web_quit_request )
    {
        size = RTSP_HEADER_LEN + MMS_HEADER_LEN;
        n = read_socket_data( &web_buf[0], RTSP_HEADER_LEN );
        if( n < RTSP_HEADER_LEN )
            return -1;
        /* Decode framing header */
        if( web_buf[0] != 0x24 )// or '$'
        {
            if( web_buf[0] < 'A' || web_buf[0] > 'z' )
            {
                DPRINT("Incorrect data\n");
                return -1;
            }
            /* We have text request or response from the server */
            DPRINT("Reading response or request from the server\n");
            init_parameters();
            if( read_response_in_pos( RTSP_HEADER_LEN, rtsp_header_line_parser ) < 0 )
                return -1;
            if( com_content_length > 0 )
            {
                WEB_MSG("Need to process content\n");
                return -1;
            }
            /* FIXME if server request is possible, processing server request
            if( com_content_length > 0 && (read_pos = skip_content( (size_t) read_pos )) < 0 )
                return -1;*/
            
            continue;
        }
        //packet_id = web_buf[1];// type of the packet, usially 0
        // if( web_buf[1] != 0 )
        // DPRINT( "id: 0x%02x\n", web_buf[1] );
        size = ((unsigned char)web_buf[2] << 8) | (unsigned char)web_buf[3];// uint16
        // buffer size must be 65526 bytes or more
        if( size < RTSP_HEADER_LEN - 4 + MMS_HEADER_LEN )
        {// framing header size - packet length position + mms header size
            DPRINT( "Incorrect chunk size: %u\n", size );
            return -1;
        }
        size -= RTSP_HEADER_LEN - 4 + MMS_HEADER_LEN;
        // int32 reason = (web_buf[4]<<24)|(web_buf[5]<<16)|(web_buf[6]<<8)|web_buf[7];
        // 80 e0 6a 6c, 80 e0 6a 6d
        /* Decode MMS header */
        n = read_socket_data( &web_buf[0], MMS_HEADER_LEN );
        if( n < MMS_HEADER_LEN )
            return -1;
        // uint32 location_id = (web_buf[0]<<24)|(web_buf[1]<<16)|(web_buf[2]<<8)|web_buf[3];
        // 37 a6 1c 46, 37 a6 1d c6
        // incarnation = web_buf[4];// ignore
        // ca
        // uint8 afflags = web_buf[5];// usially 0xfe
        // fe
        // packet_size = ((unsigned char)web_buf[6] << 8)|(unsigned char)web_buf[7];// uint16
        // d3 ad
        // payload = (web_buf[8]<<24)|(web_buf[9]<<16)|(web_buf[10]<<8)|web_buf[11];
        // 40 00 09 20
        read_pos = 0;
        do
        {
            block = size - read_pos;
            if( block > BUF_SIZE )
                block = BUF_SIZE;
            n = read_socket_data( &web_buf[0], block );
            if( out_sound_stream( n ) < 0 )
                return -1;
            read_pos += n;
            if( n != block )
                return -1;
            /* Check timeout period and send Keep Alive */
            curr_time = time( NULL );
            if( (curr_time - prev_time) > web_timeout )
            {
                if( send_keep_alive() < 0 )
                    return -1;
                prev_time = curr_time;
            }
        }
        while( read_pos < size );
    }
    return 0;
}

int rtsp()
{
    int n;

    web_cseq = 0;
    web_timeout = 60;// sec, standart timeout
    /* Check the port number */
    if( web_port <= 0 || web_port > 65535 ) web_port = RTSP_DEFAULT_PORT;
    /* Connecting to the server */
    if( connect_to_server() < 0 )
    {
        web_port = RTSP_ALT_PORT;
        if( connect_to_server() < 0 )
            return -1;
    }

    n = describe();
    if( n >= 0 )
    {
        n = play();
    }
    /* Close session */
    close_session();
    disconnect();
    WEB_MSG("Good Bye!\n");
    return n;
}
