/**
web radio client
@author vdv, mailto: ben-gan@mail.ru
*/

#include <string.h>// strlen, strstr, strncasecmp
#include <ctype.h>// isdigit, isspace
#include <stdio.h>// printf
#include <stdlib.h>// atoi
#include <signal.h>// signal

#include "common.h"
#include "http.h"
#include "rtsp.h"

/** Supported protocols */
static const char* protocol_name_tab[] = {
    "", "http://", "rtsp://" };// "mms://", "nfs://", "ftp://", "smb://" };

/**
 * Parse the url
 * @param    url - input string in format:
 * protocol://user:password@xxx.xxx.xxx.xxx:port/file
 * protocol://user:password@host:port/file
 * @return    protocol, username, password, host, port, file
 * every pointer may be NULL
 */
static void url_parser
(
    const char *url,
    const char** protocol,
    char *user,
    char *password,
    char *host,
    int *port,
    char *file
)
{
    int n;

    if( user != NULL ) *user = 0;
    if( password != NULL ) *password = 0;
    if( host != NULL ) *host = 0;
    if( port != NULL ) *port = 0;
    if( file != NULL ) *file = 0;

    while( *url == ' ' ) url++;// skip spaces

    // get protocol
    *protocol = protocol_name_tab[0];
    for( n = sizeof(protocol_name_tab)/sizeof(protocol_name_tab[0]); n--; )
    {
        if( strnicmp(url, protocol_name_tab[n], strlen(protocol_name_tab[n])) == 0 )
        {
            url += strlen( protocol_name_tab[n] );
            *protocol = protocol_name_tab[n];
            break;
        }
    }
    if( *protocol == protocol_name_tab[0] ) return;// protocol unknown

    // get user & password
    if( strstr(url, "@") != NULL )
    {
        n = MAX_USER_NAME;
        while( *url != 0 && *url != ':' )
        {
            if( user != NULL && n-- > 0 )
            {
                *user++ = *url;
                *user   = 0;
            }
            url++;
        }
        url++;// skip ':'
        n = MAX_PASSWORD_LENGTH;
        while( *url != 0 && *url != '@' )
        {
            if( password != NULL && n-- > 0 )
            {
                *password++ = *url;
                *password   = 0;
            }
            url++;
        }
        url++;// skip '@'
    }

    // get host
    n = MAX_HOST_NAME;
    while( *url != 0 && *url != ':' && *url != '/' )
    {
        if( host != NULL && n-- > 0 )
        {
            *host++ = *url;
            *host   = 0;
        }
        url++;
    }

    // get port
    if( *url == ':' )
    {
        url++;// skip ':'
        if( port != NULL )
            *port = atoi( url );
        while( isdigit(*url) ) url++;// skip port
    }

    // get file
    n = MAX_FILE_NAME;
    if( *url == '/' )
    {
        url++;// skip '/'
        while( *url != 0 && !isspace(*url) )
        {
            if( file != NULL && n-- > 0 )
            {
                *file++ = *url;
                *file   = 0;
            }
            url++;
        }
    }
}

/** Signal processing */
static void signal_handler(const int sig)
{
    switch( sig )
    {
    case SIGINT:
    case SIGPIPE:
    case SIGQUIT:
    case SIGTERM:
    case SIGKILL:
        web_quit_request = 1;
        WEB_MSG( "Caught signal %d, exiting nicely\n", sig );
        break;
    }
}
  
int main(int argc, char **argv)
{
    /* Init */
    web_quit_request = 0;
    memset( web_metadata_buf, 0, sizeof( web_metadata_buf ) );
    /* Catch terminate signals */
    signal( SIGTERM, signal_handler ); // kill
    signal( SIGHUP, signal_handler );  // kill -HUP  /  xterm closed
    signal( SIGINT, signal_handler );  // Interrupt from keyboard
    signal( SIGQUIT, signal_handler ); // Quit from keyboard
    signal( SIGPIPE, signal_handler ); // Some window managers cause this
    /* Explore parameters */
    web_proxy = NULL;
    if( argc < 2 )
    {
        printf( "Usage: %s <address> [<proxyaddr> <proxyport>]\n" \
                "address format: protocol://user:password@addr:port/file\n" \
                "supported protocol: http\n" \
                "example: web_radio http://rockradio.de:8090\n",
                argv[0] );
        exit( 1 );
    } else if( argc == 2 )
    {
        if( strncmp(argv[1], "-v", strlen("-v")) == 0 )
        {
            printf( "Web radio, version: "VERSION"\n" );
            exit( 0 );
        }
        if( strncmp(argv[1], "-h", strlen("-h")) == 0 )
        {
            printf( "Using: web_radio protocol://host:port/file\n" );
            exit( 0 );
        }
        url_parser( argv[1],// url
                    &web_protocol,// protocol
                    NULL,// user
                    NULL,// password
                    web_host,// host
                    &web_port,// port
                    web_file );// file
        DPRINT("protocol: %s\naddress: %s\nport: %d\nfile: %s\n",
                    web_protocol, web_host, web_port, web_file);
    } else if( argc == 4 )
    {
        web_proxy = argv[2];
        web_proxy_port = atoi( argv[3] );
        if( web_proxy_port <= 0 || web_proxy_port > 65535 )
        {
            fprintf( stderr, "Incorrect proxy port: %d\n", web_proxy_port );
            exit( 1 );
        }
        printf( "Using proxy %s:%d\n", web_proxy, web_proxy_port );
    }
    /* Select handler based on the protocol name */
    if( strcmp(web_protocol, "http://") == 0 )// shoutcast/icecast/wma
    {
        return http();
    } else if( strcmp(web_protocol, "rtsp://") == 0 )// rtsp(wma)
    {
        return rtsp();
    }
    printf( "Unknown protocol %s\n", web_protocol );
    return 0;
}
